/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <errno.h>
#include <glib.h>
#include <dlog.h>
#include <wayland-client.h>
#include <Ecore_Wayland.h>
#include <Elementary.h>
#include <wayland-extension/tizen-extension-client-protocol.h>
#include <tizen-remote-surface-client-protocol.h>

#include "screen_connector_trs.h"
#include "screen_connector_toolkit.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "SC_WATCHER"

struct _screen_connector_type_h {
	GHashTable *toolkit_table;
	aul_screen_viewer_h aul_handle;
};

struct _screen_connector_toolkit_h {
	void (*update_cb)(struct tizen_remote_surface *trs, struct wl_buffer *buffer, uint32_t time, void *data);
	void (*missing_cb)(struct tizen_remote_surface *trs, void *data);
	void (*change_cb)(struct tizen_remote_surface *trs, uint32_t type, struct wl_buffer *tbm,
			int32_t img_file_fd, uint32_t img_file_size, uint32_t time, struct wl_array *keys, void *data);
	screen_connector_toolkit_ops *ops;
	unsigned int event_filter;
	char *appid;
	char *instance_id;
	int pid;
	int surface_id;
	struct tizen_remote_surface *surface;
	struct _screen_connector_type_h *type_h;
	void *data;
	struct wl_surface *bind_surface;
};

static GHashTable *__type_table;
static void __buffer_updated(void *data, struct tizen_remote_surface *trs, struct wl_buffer *buffer, uint32_t time)
{
}

static void __buffer_missing(void *data, struct tizen_remote_surface *trs)
{
	screen_connector_toolkit_h toolkit_h = (screen_connector_toolkit_h)data;

	if (toolkit_h)
		toolkit_h->ops->removed_cb(toolkit_h->appid, toolkit_h->instance_id, toolkit_h->pid, toolkit_h->data);
}

static void __buffer_changed(void *data, struct tizen_remote_surface *trs, uint32_t type, struct wl_buffer *tbm,
		int32_t img_file_fd, uint32_t img_file_size, uint32_t time, struct wl_array *keys)
{
	screen_connector_toolkit_h toolkit_h = (screen_connector_toolkit_h)data;

	if (toolkit_h && toolkit_h->ops)
		toolkit_h->ops->updated_cb(trs, type, tbm, img_file_fd, img_file_size, time, keys, toolkit_h->appid,
				toolkit_h->instance_id, toolkit_h->pid, toolkit_h->data);
}

static void __input_event_filter(void *data, struct tizen_remote_surface *trs, uint32_t event_filter)
{
	screen_connector_toolkit_h toolkit_h = (screen_connector_toolkit_h)data;

	if (toolkit_h)
		toolkit_h->event_filter = event_filter;

	LOGD("event filter updated:%d", event_filter);
}

static const struct tizen_remote_surface_listener __rs_listener = {
	__buffer_updated,
	__buffer_missing,
	__buffer_changed,
	__input_event_filter,
};

EXPORT_API int screen_connector_toolkit_redirect_surface(screen_connector_toolkit_h toolkit_h)
{

	struct wl_tbm *tbm = screen_connector_get_tbm();
	struct tizen_remote_surface_manager *rsm = screen_connector_get_rsm();
	if (toolkit_h == NULL || rsm == NULL || tbm == NULL)
		return -1;

	toolkit_h->surface = tizen_remote_surface_manager_create_surface(rsm, (uint32_t)toolkit_h->surface_id, tbm);
	if (toolkit_h->surface == NULL)
		return -1;

	LOGD("resource_id : %d, (%p), %s", toolkit_h->surface_id, toolkit_h->surface, toolkit_h->instance_id);
	tizen_remote_surface_add_listener(toolkit_h->surface, &__rs_listener, toolkit_h);
	tizen_remote_surface_redirect(toolkit_h->surface);

	if (toolkit_h->bind_surface != NULL)
		tizen_remote_surface_manager_bind_surface(rsm, toolkit_h->bind_surface, toolkit_h->surface);

	return 0;
}

static void __destroy_toolkit_h(gpointer data)
{
	screen_connector_toolkit_h toolkit_h = (screen_connector_toolkit_h)data;
	if (!toolkit_h)
		return;

	LOGD("__destroy_toolkit_h appid: %s surface_id %d", toolkit_h->appid, toolkit_h->surface_id);

	if (toolkit_h->appid) {
		free(toolkit_h->appid);
		toolkit_h->appid = NULL;
	}

	if (toolkit_h->instance_id) {
		free(toolkit_h->instance_id);
		toolkit_h->instance_id = NULL;
	}

	if (toolkit_h->surface) {
		tizen_remote_surface_set_user_data(toolkit_h->surface, NULL);
		tizen_remote_surface_unredirect(toolkit_h->surface);
		tizen_remote_surface_destroy(toolkit_h->surface);
		toolkit_h->surface = NULL;
	}

	if (toolkit_h->ops) {
		free(toolkit_h->ops);
		toolkit_h->ops = NULL;
	}

	free(toolkit_h);
}

static void __destroy_type_h(gpointer data)
{
	screen_connector_type_h type_h = (screen_connector_type_h)data;
	if (!type_h)
		return;

	if (type_h->aul_handle)
			aul_screen_connector_remove_screen_viewer(type_h->aul_handle);
	if (type_h->toolkit_table) {
		g_hash_table_destroy(type_h->toolkit_table);
		type_h->toolkit_table = NULL;
	}
	free(type_h);
}

static void __aul_screen_viewer_cb(const char *appid, const char *instance_id,
		const int pid, const unsigned int surface_id,
		aul_screen_connector_event_type_e event_type, void *data)
{
	screen_connector_toolkit_h toolkit_h;
	screen_connector_type_h type_h = (screen_connector_type_h)data;

	LOGD("__aul_screen_viewer_cb pid: %d, surface_id: %d, appid: %s instance_id: %s type %d",
		pid, surface_id, appid, instance_id, event_type);

	if (type_h == NULL || type_h->toolkit_table == NULL) {
		LOGE("NULL type info");
		return;
	}

	toolkit_h = (screen_connector_toolkit_h)g_hash_table_lookup(type_h->toolkit_table, instance_id);
	if (toolkit_h == NULL) {
		LOGD("It's not my instance_id, maybe alread removed %s ", instance_id);
		return;
	}

	toolkit_h->appid = strdup(appid);
	toolkit_h->pid = pid;
	toolkit_h->surface_id = surface_id;
	if (event_type == AUL_SCREEN_CONNECTOR_EVENT_TYPE_ADD) {

		toolkit_h->ops->added_cb(appid, instance_id, pid, toolkit_h->data);
		screen_connector_toolkit_redirect_surface(toolkit_h);

	} else if (event_type == AUL_SCREEN_CONNECTOR_EVENT_TYPE_UPDATE) {

	} else if (event_type == AUL_SCREEN_CONNECTOR_EVENT_TYPE_REMOVE) {

		if (toolkit_h->ops)
			toolkit_h->ops->removed_cb(appid, instance_id, pid, toolkit_h->data);

		g_hash_table_remove(type_h->toolkit_table, instance_id);
	}
}

EXPORT_API int screen_connector_toolkit_init(screen_connector_screen_type_e type)
{
	int ret = 0;
	screen_connector_type_h type_h;

	ret = screen_connector_trs_init();
	if (__type_table == NULL)
		__type_table = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, __destroy_type_h);

	type_h = g_hash_table_lookup(__type_table, GINT_TO_POINTER(type));
	if (type_h == NULL) {

		type_h = (screen_connector_type_h)calloc(1, sizeof(struct _screen_connector_type_h));
		g_hash_table_insert(__type_table, GINT_TO_POINTER(type), type_h);
	}

	return ret;
}

EXPORT_API int screen_connector_toolkit_fini(screen_connector_screen_type_e type)
{
	int ret = 0;
	screen_connector_type_h type_h;

	if (__type_table == NULL) {
		LOGI("__type_table is NULL");
		return 0;
	}

	type_h = g_hash_table_lookup(__type_table, GINT_TO_POINTER(type));
	if (type_h)
		g_hash_table_remove(__type_table, GINT_TO_POINTER(type));

	if (g_hash_table_size(__type_table) == 0) {
		ret = screen_connector_trs_fini();
		g_hash_table_destroy(__type_table);
		__type_table = NULL;
	}

	return ret;
}

EXPORT_API screen_connector_toolkit_h screen_connector_toolkit_create_handle(char *id, int surface_id,
	screen_connector_toolkit_ops *ops, void *data)
{
	screen_connector_toolkit_h toolkit_h = NULL;
	screen_connector_toolkit_ops *ops_copy;

	ops_copy = (screen_connector_toolkit_ops *)calloc(1, sizeof(screen_connector_toolkit_ops));
	if (ops_copy == NULL) {
		LOGE("ops_copy calloc fail");
		return NULL;
	}
	memcpy(ops_copy, ops, sizeof(screen_connector_toolkit_ops));

	toolkit_h = (screen_connector_toolkit_h)calloc(1, sizeof(struct _screen_connector_toolkit_h));
	if (toolkit_h == NULL) {
		LOGE("toolkit_h calloc fail");
		free(ops_copy);
		return NULL;
	}
	toolkit_h->surface_id = surface_id;
	toolkit_h->appid = strdup(id);
	toolkit_h->instance_id = strdup(id);
	toolkit_h->data = data;
	toolkit_h->ops = ops_copy;

	return toolkit_h;
}

EXPORT_API screen_connector_toolkit_h screen_connector_toolkit_add_with_surface(screen_connector_toolkit_ops *ops, char *id,
	screen_connector_screen_type_e type, struct wl_surface *surface, void *data)
{
	screen_connector_toolkit_h toolkit_h;
	screen_connector_type_h type_h;
	screen_connector_toolkit_ops *ops_copy;

	if (id == NULL || ops == NULL) {
		LOGE("Invalid param");
		return NULL;
	}

	type_h = g_hash_table_lookup(__type_table, GINT_TO_POINTER(type));
	if (type_h == NULL) {
		LOGE("type %d is not initialized !!", type);
		return NULL;
	}

	ops_copy = (screen_connector_toolkit_ops *)calloc(1, sizeof(screen_connector_toolkit_ops));
	if (ops_copy == NULL) {
		LOGE("ops_copy calloc fail");
		return NULL;
	}
	memcpy(ops_copy, ops, sizeof(screen_connector_toolkit_ops));

	toolkit_h = (screen_connector_toolkit_h)calloc(1, sizeof(struct _screen_connector_toolkit_h));
	if (toolkit_h == NULL) {
		LOGE("toolkit_h calloc fail");
		free(ops_copy);
		return NULL;
	}

	toolkit_h->instance_id = strdup(id);
	if (toolkit_h->instance_id == NULL) {
		LOGE("strdup id fail");
		free(ops_copy);
		free(toolkit_h);
		return NULL;
	}

	toolkit_h->data = data;
	toolkit_h->ops = ops_copy;
	toolkit_h->type_h = type_h;
	toolkit_h->bind_surface = surface;

	if (type_h->toolkit_table == NULL)
		type_h->toolkit_table = g_hash_table_new_full(g_str_hash,
				g_str_equal, free, __destroy_toolkit_h);

	g_hash_table_insert(type_h->toolkit_table, strdup(id), toolkit_h);

	if (type_h->aul_handle == NULL)
		aul_screen_connector_add_screen_viewer(__aul_screen_viewer_cb, type, true, type_h, &type_h->aul_handle);

	return toolkit_h;
}

EXPORT_API screen_connector_toolkit_h screen_connector_toolkit_add(screen_connector_toolkit_ops *ops, char *id,
	screen_connector_screen_type_e type, void *data)
{
	screen_connector_toolkit_h toolkit_h = NULL;
	screen_connector_type_h type_h = NULL;
	screen_connector_toolkit_ops *ops_copy;

	if (id == NULL || ops == NULL) {
		LOGE("Invalid param");
		return NULL;
	}

	type_h = g_hash_table_lookup(__type_table, GINT_TO_POINTER(type));
	if (type_h == NULL) {
		LOGE("type %d is not initialized !!", type);
		return NULL;
	}

	ops_copy = (screen_connector_toolkit_ops *)calloc(1, sizeof(screen_connector_toolkit_ops));
	if (ops_copy == NULL) {
		LOGE("ops_copy calloc fail");
		return NULL;
	}
	memcpy(ops_copy, ops, sizeof(screen_connector_toolkit_ops));

	toolkit_h = (screen_connector_toolkit_h)calloc(1, sizeof(struct _screen_connector_toolkit_h));
	if (toolkit_h == NULL) {
		LOGE("toolkit_h calloc fail");
		free(ops_copy);
		return NULL;
	}

	toolkit_h->instance_id = strdup(id);
	toolkit_h->data = data;
	toolkit_h->ops = ops_copy;
	toolkit_h->type_h = type_h;

	if (type_h->toolkit_table == NULL)
		type_h->toolkit_table = g_hash_table_new_full(g_str_hash,
				g_str_equal, free, __destroy_toolkit_h);

	g_hash_table_insert(type_h->toolkit_table, strdup(id), toolkit_h);

	if (type_h->aul_handle == NULL)
		aul_screen_connector_add_screen_viewer(__aul_screen_viewer_cb, type, true, type_h, &type_h->aul_handle);

	return toolkit_h;
}

EXPORT_API int screen_connector_toolkit_remove(screen_connector_toolkit_h h)
{
	if (h == NULL) {
		LOGE("screen_connector_toolkit_h is NULL");
		return -1;
	}

	if (h->type_h == NULL) {
		LOGE("screen_connector_toolkit_h type_h is NULL");
		return -1;
	}

	if (h->type_h->toolkit_table == NULL) {
		LOGE("screen_connector_toolkit_h type_h toolkit_table is NULL");
		return -1;
	}

	g_hash_table_remove(h->type_h->toolkit_table, h->instance_id);

	return 0;
}

EXPORT_API int screen_connector_toolkit_update(const char *appid, const char *instance_id)
{
	return aul_screen_connector_send_update_request(appid, instance_id);
}

EXPORT_API const char *screen_connector_toolkit_get_appid(screen_connector_toolkit_h h)
{
	if (h == NULL)
		return NULL;

	return h->appid;
}

EXPORT_API const char *screen_connector_toolkit_get_instance_id(screen_connector_toolkit_h h)
{
	if (h == NULL)
		return NULL;

	return h->instance_id;
}

EXPORT_API int screen_connector_toolkit_get_pid(screen_connector_toolkit_h h)
{
	if (h == NULL)
		return -1;

	return h->pid;
}

EXPORT_API int screen_connector_toolkit_get_surface_id(screen_connector_toolkit_h h)
{
	if (h == NULL)
		return -1;

	return h->surface_id;
}

EXPORT_API void *screen_connector_toolkit_get_user_data(screen_connector_toolkit_h h)
{
	if (h == NULL)
		return NULL;

	return h->data;
}

EXPORT_API unsigned int screen_connector_toolkit_get_event_filter(screen_connector_toolkit_h h)
{
	if (h == NULL)
		return 0;

	return h->event_filter;
}

EXPORT_API int screen_connector_toolkit_check_input_disabled(screen_connector_toolkit_h h,
	screen_connector_input_type_e input)
{
	if (h == NULL) {
		LOGE("inavlid argument");
		return -1;
	}

	return ((h->event_filter & input) > 0);
}

EXPORT_API struct tizen_remote_surface *screen_connector_toolkit_get_trs(screen_connector_toolkit_h h)
{
	if (h == NULL)
		return 0;

	return h->surface;
}
